// -----------------------------------------------------------------------------
// SWIG Interface to Clips
//
// Dave Beazley (beazley@cs.uchicago.edu)
// University of Chicago
// August 4, 1999
// -----------------------------------------------------------------------------

%title "Clips", sort, pre

%module clips
%init %{
if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
       return TCL_ERROR;
}
%}
%{
#undef SWIG_prefix
#define SWIG_prefix "clips::"
%}
 
%{
#include "clips.h"
#include "strngfun.h"

  /* Prototypes missing from headers */
LOCALE long int                       FactExistp(void *);
LOCALE void                           FactSlotNames(void *,DATA_OBJECT *);
LOCALE void                           GetFactList(DATA_OBJECT *,void *);
LOCALE void *GetNextInstanceInClassAndSubclasses(void **,void *,DATA_OBJECT *);

%}

%disabledoc

// -----------------------------------------------------------------------------
// Typemaps and interface configuration
// -----------------------------------------------------------------------------
// Map constants to symbolic names
#ifdef SWIGTCL

%{
static Tcl_HashTable  constTable;      /* Hash table          */
static int           *swigconst;       /* Temporary variable  */
static Tcl_HashEntry *entryPtr;        /* Hash entry          */
static int            dummy;           /* dummy value         */
%}

// Allow symbolic constants to be passed as integers */

%init %{
  Tcl_InitHashTable(&constTable,TCL_STRING_KEYS);
%}

%typemap(tcl,const) int, unsigned int, long, unsigned long {
  entryPtr = Tcl_CreateHashEntry(&constTable,"$target",&dummy);
  swigconst = (int *) malloc(sizeof(int));
  *swigconst = $source;
  Tcl_SetHashValue(entryPtr, swigconst);
  /* Make it so constants can also be used as variables */
  Tcl_LinkVar(interp, SWIG_prefix "$target", (char *) swigconst, TCL_LINK_INT | TCL_LINK_READ_ONLY);
};

// Now change integer handling hooks to look for names
%typemap(tcl,in) int, unsigned int, long, unsigned long {
  Tcl_HashEntry *entryPtr;        
  entryPtr = Tcl_FindHashEntry(&constTable,$source);
  if (entryPtr) {
    $target = ($type) (*((int *) Tcl_GetHashValue(entryPtr)));
  } else {
    $target = ($type) atoi($source);
  }
}

// Map the string "NULL" to a NULL char *
%typemap(tcl,check) char *namenull {
  if (strcmp($target,"NULL") == 0) {
    $target = 0;
  }
}

%inline %{
int  NameToInt(int a) { return a; }
%}

#endif


// -----------------------------------------------------------------------------
// Magic typemaps for resultbuffer, resultlen combinations
// -----------------------------------------------------------------------------

%typemap(tcl,ignore) char *resultbuffer(char **rbuf) {
    $target = 0;
    rbuf = &$target;
}

%typemap(tcl,argout) char *resultbuffer {
  Tcl_SetResult(interp,$source, TCL_VOLATILE);
  free($source);
}

%typemap(tcl,check) int resultlen {
  if ($target > 0) {
    *rbuf = (char *) malloc($target);
    if (!(*rbuf)) {
      Tcl_SetResult(interp,"Fatal Error.  No memory available for return result!", TCL_STATIC);
      return TCL_ERROR;
    }
  } else {
    *rbuf = (char *) malloc(16);
    $target = 0;
  }
}

// -----------------------------------------------------------------------------
// Default handling of modules
// -----------------------------------------------------------------------------

%typemap(tcl,default) void *theModule {
    $target = GetCurrentModule();
}
%typemap(tcl,in) void *theModule, void *module {
  if (strcmp($source,"current")) {
    if (SWIG_GetPtr($source,&$target,"_void_p")) {
      Tcl_SetResult(interp,"Type error. Expected _void_p.",TCL_STATIC);
      return TCL_ERROR;
    }
  } else {
    $target = GetCurrentModule();
  }
}

%typemap(tcl,ignore) Tcl_Interp * {$target = interp;}

typedef int BOOLEAN;
%typedef int tclint;
%typedef int typeint;

%typemap(tcl,out) typeint {
    char *tname;
    switch($source) {
    case STRING:
      tname = "STRING";
      break;
    case INTEGER:
      tname = "INTEGER";
      break;
    case SYMBOL:
      tname = "SYMBOL";
      break;
    case FLOAT:
      tname = "FLOAT";
      break;
    case EXTERNAL_ADDRESS:
      tname = "EXTERNAL_ADDRESS";
      break;
    case INSTANCE_ADDRESS:
      tname = "INSTANCE_ADDRESS";
      break;
    case INSTANCE_NAME:
      tname = "INSTANCE_NAME";
      break;
    case FACT_ADDRESS:
      tname = "FACT_ADDRESS";
      break;
    case MULTIFIELD:
      tname = "MULTIFIELD";
      break;
    default :
      tname = "UNKNOWN";
      break;
    }
    Tcl_SetResult(interp,tname,TCL_STATIC);
}
      
%typemap(tcl,out) tclint { return $source; }

%enabledoc

%section "CLIPS Constants",sort
%text %{
The following constants are used by some of the CLIPS functions.
Each constant is accessed in two ways:

    1.  As a symbolic name such as 'STRING'
    2.  By value such as '$clips::STRING'

  %}

// -----------------------------------------------------------------------------
// Useful constants 
// -----------------------------------------------------------------------------

enum { FALSE, TRUE, EXACTLY, AT_LEAST, NO_MORE_THAN, RANGE };

enum {FLOAT, INTEGER, SYMBOL, STRING, MULTIFIELD, EXTERNAL_ADDRESS, FACT_ADDRESS,
      INSTANCE_ADDRESS, INSTANCE_NAME };

enum { SYMBOL_OR_STRING, INTEGER_OR_FLOAT, INSTANCE_OR_INSTANCE_NAME };

%section "DataObject"

%text %{
The following functions are used to manpulate variables of type
DATA_OBJECT and its associated values.
%}

// -----------------------------------------------------------------------------
// DataObject methods
// -----------------------------------------------------------------------------

%inline %{
DATA_OBJECT *CreateDataObject() {
    return (DATA_OBJECT *) malloc(sizeof(DATA_OBJECT));
}
 void DestroyDataObject(DATA_OBJECT *d) {
   free(d);
 }

%}

%name(GetType)     typeint GetpType(DATA_OBJECT *);
%name(DOToDouble)  double DOPToDouble(DATA_OBJECT *);
%name(DOToInteger) int    DOPToInteger(DATA_OBJECT *);
%name(DOToString)  char  *DOPToString(DATA_OBJECT *);
%name(DOToLong)    long   DOPToLong(DATA_OBJECT *);
%name(DOToFloat)   float  DOPToFloat(DATA_OBJECT *);
%name(DOToPointer) void  *DOPToPointer(DATA_OBJECT *);

%inline %{
int IsSTRING(DATA_OBJECT *dobj) {
  return GetpType(dobj) == STRING ? 1 : 0;
}

int IsINTEGER(DATA_OBJECT *dobj) {
  return GetpType(dobj) == INTEGER ? 1 : 0;
}
int IsFLOAT(DATA_OBJECT *dobj) {
  return GetpType(dobj) == FLOAT ? 1 : 0;
}
int IsSYMBOL(DATA_OBJECT *dobj) {
  return GetpType(dobj) == SYMBOL ? 1 : 0;
}
int IsINSTANCE_NAME(DATA_OBJECT *dobj) {
  return GetpType(dobj) == INSTANCE_NAME ? 1 : 0;
}
int IsINSTANCE_ADDRESS(DATA_OBJECT *dobj) {
  return GetpType(dobj) == INSTANCE_ADDRESS ? 1 : 0;
}

int IsEXTERNAL_ADDRESS(DATA_OBJECT *dobj) {
  return GetpType(dobj) == EXTERNAL_ADDRESS ? 1 : 0;
}

int IsFACT_ADDRESS(DATA_OBJECT *dobj) {
  return GetpType(dobj) == FACT_ADDRESS ? 1 : 0;
}

int IsMULTIFIELD(DATA_OBJECT *dobj) {
  return GetpType(dobj) == MULTIFIELD ? 1 : 0;
}
 %}

// -----------------------------------------------------------------------------
// Macros for extracting values.
// -----------------------------------------------------------------------------

%name(GetValue)    void  *GetpValue(DATA_OBJECT *);
char     *ValueToString(void *value);
double    ValueToDouble(void *value);
long      ValueToLong(void *value);
int       ValueToInteger(void *value);

// Magic DataObject value extractor.   Returns the right value from a data
// object dynamically
%inline %{
tclint GetAnyValue(Tcl_Interp *interp, DATA_OBJECT *d) {
  int type;
  char stemp[256];
  type = GetpType(d);
  switch(type) {
  case STRING:
  case SYMBOL:
  case INSTANCE_NAME:
    Tcl_SetResult(interp, DOPToString(d),TCL_VOLATILE);
    break;
  case FLOAT:
    Tcl_PrintDouble(interp, DOPToDouble(d), stemp);
    Tcl_SetResult(interp,stemp,TCL_VOLATILE);
    break;
  case INTEGER:
    sprintf(stemp,"%d",DOPToInteger(d));
    Tcl_SetResult(interp,stemp,TCL_VOLATILE);
    break;
  case MULTIFIELD:
    SWIG_MakePtr(stemp, GetpValue(d),"_void_p");
    Tcl_SetResult(interp,stemp,TCL_VOLATILE);    
    break;
  case INSTANCE_ADDRESS:
  case EXTERNAL_ADDRESS:
  case FACT_ADDRESS:
    SWIG_MakePtr(stemp, DOPToPointer(d),"_void_p");
    Tcl_SetResult(interp,stemp,TCL_VOLATILE);    
    break;
  default:
    break;
  }
  return TCL_OK;
}
%}

// -----------------------------------------------------------------------------
// Macros for setting values
// -----------------------------------------------------------------------------

%name(SetType) int SetpType(DATA_OBJECT *, int type);
%name(SetValue) int SetpValue(DATA_OBJECT *, void *value);
void *AddLong(long );
void *AddDouble(double);
void *AddSymbol(char *);

%section "Multifields"

// -----------------------------------------------------------------------------
// Multifield values
// -----------------------------------------------------------------------------

%name(GetDOLength) int   GetpDOLength(DATA_OBJECT *);
%name(GetDOBegin)  int   GetpDOBegin(DATA_OBJECT *);
%name(GetDOEnd)    int   GetpDOEnd(DATA_OBJECT *);
%name(SetDOBegin) int      SetpDOBegin(DATA_OBJECT *value,int position);
%name(SetDOEnd)   int      SetpDOEnd(DATA_OBJECT *value,int position);
void              *CreateMultifield(int size);


typeint       GetMFType(void *multifieldPtr, int position);
%name(GetMFTypeInt) int           GetMFType(void *multifieldPtr, int position);
void         *GetMFValue(void *multifieldPtr, int position);
int           SetMFType(void *multifield, int position, int type);
void         *SetMFValue(void *multifield, int position, void *value);

// Magic Multifield  value extractor.   Returns the right value from a data
// object dynamically

%inline %{
  tclint GetMFAnyValue(Tcl_Interp *interp, void *mf, int position) {
    int type;
    char stemp[256];
    void *value;
    type = GetMFType(mf,position);
    value = GetMFValue(mf,position);
    switch(type) {
    case STRING:
    case SYMBOL:
    case INSTANCE_NAME:
      Tcl_SetResult(interp, ValueToString(value),TCL_VOLATILE);
      break;
    case FLOAT:
      Tcl_PrintDouble(interp, ValueToDouble(value), stemp);
      Tcl_SetResult(interp,stemp,TCL_VOLATILE);
      break;
    case INTEGER:
      sprintf(stemp,"%ld",ValueToLong(value));
      Tcl_SetResult(interp,stemp,TCL_VOLATILE);
      break;
    case MULTIFIELD:
      SWIG_MakePtr(stemp, value,"_void_p");
      Tcl_SetResult(interp,stemp,TCL_VOLATILE);    
      break;
    case INSTANCE_ADDRESS:
    case EXTERNAL_ADDRESS:
    case FACT_ADDRESS:
      SWIG_MakePtr(stemp, value,"_void_p");
      Tcl_SetResult(interp,stemp,TCL_VOLATILE);    
      break;
    default:
      break;
    }
    return TCL_OK;
  }

%}

void SetMultifieldErrorValue(DATA_OBJECT *value);

// -----------------------------------------------------------------------------
// Environment
// -----------------------------------------------------------------------------

%section "Environment"

%typedef void (*ClearFunction)();

#ifdef SWIGTCL
%{
static  char *clearproc = 0;
static  Tcl_Interp *clear_interp = 0;
void tcl_clearproc(void) {
  if (clearproc && clear_interp) {
      Tcl_GlobalEval(clear_interp,clearproc);
  }
}

int MyAddClearFunction(Tcl_Interp *interp, char *clearfunc, int priority) {
    if (clearproc) {
        RemoveClearFunction("tcl");
	free(clearproc);
    }
    clearproc = (char *) malloc(strlen(clearfunc)+8);
    strcpy(clearproc,clearfunc);
    clear_interp = interp;
    return AddClearFunction("tcl",tcl_clearproc,priority);
}
%}

%name(AddClearFunction) int MyAddClearFunction(Tcl_Interp *interp, char *clearfunc, int priority);
#endif

%typedef void (*PeriodicFunction)();
#ifdef SWIGTCL
%{
static  char *periodicproc = 0;
static  Tcl_Interp *periodic_interp = 0;
void tcl_periodicproc(void) {
  if (periodicproc && periodic_interp) {
      Tcl_GlobalEval(periodic_interp,periodicproc);
  }
}

int MyAddPeriodicFunction(Tcl_Interp *interp, char *periodicfunc, int priority) {
    if (periodicproc) {
        RemovePeriodicFunction("tcl");
    }
    free(periodicproc);
    periodicproc = (char *) malloc(strlen(periodicfunc)+8);
    strcpy(periodicproc,periodicfunc);
    periodic_interp = interp;
    return AddPeriodicFunction("tcl",tcl_periodicproc,priority);
}
%}

%name(AddPeriodicFunction) int MyAddPeriodicFunction(Tcl_Interp *interp, char *periodicfunc, int priority);
#endif

%typedef void (*ResetFunction)();
#ifdef SWIGTCL
%{
static  char *resetproc = 0;
static  Tcl_Interp *reset_interp = 0;
void tcl_resetproc(void) {
  if (resetproc && reset_interp) {
      Tcl_GlobalEval(reset_interp,resetproc);
  }
}

int MyAddResetFunction(Tcl_Interp *interp, char *resetfunc, int priority) {
    if (resetproc) {
        RemoveResetFunction("tcl");
    }
    free(resetproc);
    resetproc = (char *) malloc(strlen(resetfunc)+8);
    strcpy(resetproc,resetfunc);
    reset_interp = interp;
    return AddResetFunction("tcl",tcl_resetproc,priority);
}
%}

%name(AddResetFunction) int MyAddResetFunction(Tcl_Interp *interp, char *resetfunc, int priority);
#endif

int     Bload(char *filename);
int     Bsave(char *filename);
void    Clear();
void    FunctionCall(char *funcname, char *args, DATA_OBJECT *result);
int     GetAutoFloatDividend();
int     GetDynamicConstraintChecking();
int     GetSequenceOperatorRecognition();
int     GetStaticConstraintChecking();
void    InitializeEnvironment();
int     Load(char *filename);
int     RemoveClearFunction(char *clearitemname);
int     RemovePeriodicFunction(char *periodicitemname);
int     RemoveResetFunction(char *resetitemname);
void    Reset();
int     Save(char *filename);
int     SetAutoFloatDividend(int boolvalue);
int     SetDynamicConstraintChecking(int boolvalue);
int     SetSequenceOperatorRecognition(int boolvalue);
int     SetStaticConstraintChecking(int boolvalue);
int     BatchStar(char *filename);
int     Build(char *constructstring);

%typemap(tcl,default) DATA_OBJECT *evalresult (DATA_OBJECT temp) {
  $target = &temp;
}

int     Eval(char *expressionstring, DATA_OBJECT *evalresult);

// -----------------------------------------------------------------------------
// Debugging
// -----------------------------------------------------------------------------

%section "Debugging"

int     DribbleActive();
int     DribbleOff();
int     DribbleOn(char *filename);
int     GetWatchItem(char *item);
int     Unwatch(char *item);
int     Watch(char *item);

// ----------------------------------------------------------------------
// DefTemplate Functions
// -----------------------------------------------------------------------------

%section "Deftemplate Functions"

char *DeftemplateModule(void *deftemplatePtr);
void *FindDeftemplate(char *deftemplateName);
void  GetDeftemplateList(DATA_OBJECT *result, void *theModule);
char *GetDeftemplateName(void *deftemplatePtr);
char *GetDeftemplatePPForm(void *deftemplatePtr);
int   GetDeftemplateWatch(void *deftemplatePtr);
void *GetNextDeftemplate(void *deftemplatePtr);
int   IsDeftemplateDeletable(void *deftemplatePtr);
void  ListDeftemplates(char *logicalname, void *theModule);
void  SetDeftemplateWatch(int newstate, void *deftemplatePtr);
int   Undeftemplate(void *deftemplatePtr);

// -----------------------------------------------------------------------------
// Fact Functions
// -----------------------------------------------------------------------------

%section "Fact Functions"

void                *Assert(void *factPtr);
void                *AssertString(char *);
BOOLEAN              AssignFactSlotDefaults(void *);
struct fact         *CreateFact(void *deftemplatePtr);
void                 DecrementFactCount(void *);
long                 FactIndex(void *factPtr);
void                 Facts(char *logicalname,void *theModule,long start = -1,long end = -1 ,long max = -1);
int                  GetFactDuplication();
int                  GetFactListChanged();
void                 GetFactPPForm(char *resultbuffer,int resultlen,void *fact);
BOOLEAN              GetFactSlot(void *,char *namenull,DATA_OBJECT *value);
long                 GetNumberOfFacts(void);
void                *GetNextFact(void *factPtr);
void                 IncrementFactCount(void *);
int                  LoadFacts(char *filename);
int                  Retract(void *factptr);
int                  SaveFacts(char *,int,struct expr *);
int                  SetFactDuplication(int boolvalue);
void                 SetFactListChanged(int boolvalue);
long                 FactExistp(void *factptr);
void                 FactSlotNames(void *factptr, DATA_OBJECT *result);
void                 GetFactList(DATA_OBJECT *result, void *theModule);
int                  LoadFactsFromString(char *,int);
BOOLEAN              PutFactSlot(void *,char *,DATA_OBJECT *);


// -----------------------------------------------------------------------------
// Deffacts Functions
// -----------------------------------------------------------------------------

%section "Deffacts Functions"

char                *DeffactsModule(void *theDeffacts);
void                *FindDeffacts(char *deffactsName);
void                 GetDeffactsList(DATA_OBJECT *result, void *theModule);
char                *GetDeffactsName(void *deffactsPtr);
char                *GetDeffactsPPForm(void *deffactsptr);
void                *GetNextDeffacts(void *);
BOOLEAN              IsDeffactsDeletable(void *);
void                 ListDeffacts(char *logicalname, void *theModule);
int                  Undeffacts(void *deffactsPtr);

// -----------------------------------------------------------------------------
// Defrule Functions
// -----------------------------------------------------------------------------

%section "Defrule Functions"

BOOLEAN  DefruleHasBreakpoint(void *);
char    *DefruleModule(void *theDefrule);
void    *FindDefrule(char *);
void     GetDefruleList(DATA_OBJECT *result, void *theModule);
char    *GetDefruleName(void *defrule);
char    *GetDefrulePPForm(void *defrule);
int      GetDefruleWatchActivations(void *defrule);
int      GetDefruleWatchFirings(void *defrule);
int      GetIncrementalReset();
void    *GetNextDefrule(void *defrule);
int      IsDefruleDeletable(void *defrule);
void     ListDefrules(char *name, void *theModule);
int      Matches(void *defrule);
int      Refresh(void *defrule);
int      RemoveBreak(void *defrule);
void     SetBreak(void *defrule);
void     SetDefruleWatchActivations(int, void *defrule);
void     SetDefruleWatchFirings(int, void *defrule);
int      SetIncrementalReset(int boolvalue);
void     ShowBreaks(char *logname, void *theModule);
int      Undefrule(void *defrule);

// -----------------------------------------------------------------------------
// Agenda Functions
// -----------------------------------------------------------------------------

%section "Agenda Functions"

%typedef void (*RunFunction)();
#ifdef SWIGTCL
%{
static  char *runproc = 0;
static  Tcl_Interp *run_interp = 0;
void tcl_runproc(void) {
  if (runproc && run_interp) {
      Tcl_GlobalEval(run_interp,runproc);
  }
}

int MyAddRunFunction(Tcl_Interp *interp, char *runfunc, int priority) {
    if (runproc) {
      RemoveRunFunction("tcl");
    }
    free(runproc);
    runproc = (char *) malloc(strlen(runfunc)+8);
    strcpy(runproc,runfunc);
    run_interp = interp;
    return AddRunFunction("tcl",tcl_runproc,priority);
}
%}

%name(AddRunFunction) int MyAddRunFunction(Tcl_Interp *interp, char *runfunc, int priority);
#endif

void     Agenda(char *logname,void *theModule);
void     ClearFocusStack();
BOOLEAN  DeleteActivation(void *activation);
void     Focus(void *defmodule);
char    *GetActivationName(void *activation);
void     GetActivationPPForm(char *resultbuffer,int resultlen,void *activation);
int      GetActivationSalience(void *activation);
int      GetAgendaChanged();
%name(GetFocus) void *WRGetFocus();
void     GetFocusStack(DATA_OBJECT *result);
void    *GetNextActivation(void *activation);
BOOLEAN  GetSalienceEvaluation(void);
int      GetStrategy();
void     ListFocusStack(char *logicalname);
void    *PopFocus();
void     RefreshAgenda(void *theModule);
int      RemoveRunFunction(char *name);
void     ReorderAgenda(void *theModule);
long     Run(int runlimit);
int      SetActivationSalience(void *activation,int salience);
void     SetAgendaChanged(int boolvalue);
BOOLEAN  SetSalienceEvaluation(int saliencevalue);

enum  {WHEN_DEFINED, WHEN_ACTIVATED, EVERY_CYCLE};

int      SetStrategy(int strategyvalue);

enum {DEPTH_STRATEGY, BREADTH_STRATEGY, LEX_STRATEGY, MEA_STRATEGY,
      COMPLEXITY_STRATEGY, SIMPLICITY_STRATEGY, RANDOM_STRATEGY };

// -----------------------------------------------------------------------------
// Defglobal functions
// -----------------------------------------------------------------------------

%section "Defglobal Functions"

char    *DefglobalModule(void *defglobal);
void    *FindDefglobal(char *name);
void     GetDefglobalList(DATA_OBJECT *result, void *theModule);
char    *GetDefglobalName(void *defglobal);
char    *GetDefglobalPPForm(void *defglobal);
BOOLEAN  GetDefglobalValue(char *defglobal, DATA_OBJECT *result);
void     GetDefglobalValueForm(char *resultbuffer,int resultlen,void *defglobal);
int      GetDefglobalWatch(void *defglobalPtr);
int      GetGlobalsChanged();
void    *GetNextDefglobal(void *defglobal);
int      GetResetGlobals();
int      IsDefglobalDeletable(void *defglobal);
void     ListDefglobals(char *logicalname, void *theModule);
BOOLEAN  SetDefglobalValue(char *name,DATA_OBJECT *value);
void     SetGlobalsChanged(int boolvalue);
void     SetDefglobalWatch(int newstate, void *defglobal);
int      SetResetGlobals(int boolvalue);
void     ShowDefglobals(char *logname, void *theModule);
int      Undefglobal(void *defglobal);

// -----------------------------------------------------------------------------
// Deffunction Functions
// -----------------------------------------------------------------------------

%section "Deffunction Functions"

char *DeffunctionModule(void *deffunction);
void *FindDeffunction(char *);
void  GetDeffunctionList(DATA_OBJECT *result, void *theModule);
char *GetDeffunctionName(void *deffunc);
char *GetDeffunctionPPForm(void *deffunc);
int   GetDeffunctionWatch(void *deffunc);
void *GetNextDeffunction(void *);
int  IsDeffunctionDeletable(void *);
void  ListDeffunctions(char *logname, void *theModule);
void  SetDeffunctionWatch(int newstate, void *deffunc);
BOOLEAN Undeffunction(void *);


// -----------------------------------------------------------------------------
// DefGeneric Functions
// -----------------------------------------------------------------------------

%section "Defgeneric Functions"

char *DefgenericModule(void *defgeneric);
void *FindDefgeneric(char *);
void  GetDefgenericList(DATA_OBJECT *result, void *theModule);
char *GetDefgenericName(void *defgen);
char *GetDefgenericPPForm(void *defgen);
int   GetDefgenericWatch(void *defgen);
void *GetNextDefgeneric(void *defgen);
int  IsDefgenericDeletable(void *);
void  ListDefgenerics(char *name, void *theModule);
void  SetDefgenericWatch(int newstate, void *defgen);
BOOLEAN Undefgeneric(void *);

// -----------------------------------------------------------------------------
// Defmethod functions
// -----------------------------------------------------------------------------

%section "Defmethod functions"

void     GetDefmethodDescription(char *resultbuffer, int resultlen, void *defgenericPtr, int methodindex);
void     GetDefmethodList(void *defgen, DATA_OBJECT *result);
char    *GetDefmethodPPForm(void *defgen, unsigned methodindex);
char    *GetDefmethodWatch(void *defgen, unsigned methodindex);
void     GetMethodRestrictions(void *defgen,unsigned methodindex,DATA_OBJECT *result);
unsigned GetNextDefmethod(void *defgen, unsigned methodindex);
int      IsDefmethodDeletable(void *defgen, unsigned methodindex);
void     ListDefmethods(char *name, void *defgen);
void     SetDefmethodWatch(int newsate, void *defgen, unsigned methodindex);
int      Undefmethod(void *defgen, unsigned methodindex);

// -----------------------------------------------------------------------------
// Defclass functions
// -----------------------------------------------------------------------------

%section "Defclass functions"

void BrowseClasses(char *name,void *defclass);
void DescribeClass(char *name,void *defclass);
int  ClassAbstractP(void *defclass);
int  ClassReactiveP(void *defclass);
void ClassSlots(void *defclass,DATA_OBJECT *result,int inheritflag);
void ClassSubclasses(void *defclass,DATA_OBJECT *result,int inheritflag);
void ClassSuperclasses(void *defclass,DATA_OBJECT *result,int inheritflag);
char *DefclassModule(void *defclass);
void *FindDefclass(char *name);
void  GetDefclassList(DATA_OBJECT *result, void *theModule);
char *GetDefclassName(void *defclass);
char *GetDefclassPPForm(void *defclass);
int   GetDefclassWatchInstances(void *defclass);
int   GetDefclassWatchSlots(void *defclass);
void *GetNextDefclass(void *defclass);
int   IsDefclassDeletable(void *defclass);
void  ListDefclasses(char *name, void *theModule);
void  SetDefclassWatchInstances(int newstate, void *defclass);
void  SetDefclassWatchSlots(int newstate, void *defclass);
void  SlotAllowedValues(void *defclass,char *slotname,DATA_OBJECT *result);
void  SlotCardinality(void *defclass,char *slotname,DATA_OBJECT *result);
BOOLEAN SlotDirectAccessP(void *defclass,char *slotname);
BOOLEAN SlotExistP(void *defclass,char *slotname,BOOLEAN inherit);
void SlotFacets(void *defclass,char *slotname,DATA_OBJECT *result);
BOOLEAN SlotInitableP(void *defclass,char *slotname);
BOOLEAN SlotPublicP(void *defclass,char *slotname);
void SlotRange(void *defclass,char *slotname,DATA_OBJECT *result);
void SlotSources(void *defclass,char *slotname,DATA_OBJECT *result);
void SlotTypes(void *defclass,char *slotname,DATA_OBJECT *result);
BOOLEAN SlotWritableP(void *defclass,char *slotname);
BOOLEAN SubclassP(void *defclass1,void *defclass2);
BOOLEAN SuperclassP(void *defclass1,void *defclass2);
int     Undefclass(void *defclass);

// -----------------------------------------------------------------------------
// Instance Functions
// -----------------------------------------------------------------------------

%section "Instance Functions"

long BinaryLoadInstances(char *filename);
long BinarySaveInstances(char *filename ,int savecode, EXPRESSION *expr = 0, BOOLEAN b = 1);

enum {LOCAL_SAVE, VISIBLE_SAVE};

void *CreateRawInstance(void *defclass,char *instancename);
void  DecrementInstanceCount(void *instance);
BOOLEAN DeleteInstance(void *instance);
void DirectGetSlot(void *instance,char *slotname,DATA_OBJECT *result);
int DirectPutSlot(void *instance,char *slotname,DATA_OBJECT *value);
void *FindInstance(void *module,char *instancename,BOOLEAN searchimports);
void *GetInstanceClass(void *instance);
char *GetInstanceName(void *instance);
void  GetInstancePPForm(char *resultbuffer, int resultlen, void *instance);
int   GetInstancesChanged();
void *GetNextInstance(void *instance);
void *GetNextInstanceInClass(void *defclass,void *instance);
void *GetNextInstanceInClassAndSubclasses(void **defclassptr,void *instance,DATA_OBJECT *result);
void  IncrementInstanceCount(void *instance);
void  Instances(char *logname,void *module,char *classname,int subclassflag);
long  LoadInstances(char *filename);
void *MakeInstance(char *makecommand);
long  RestoreInstances(char *filename);
long  SaveInstances(char *filename, int savecode, void *e = 0, int b = 0);
void  Send(DATA_OBJECT *value, char *msg, char *msgargs, DATA_OBJECT *result);
void  SetInstancesChanged(int changedflag);
int   UnmakeInstance(void *instance);
int   ValidInstanceAddress(void *instance);
int   LoadInstancesFromString(char *inputstring, int maxposition = -1);
int   RestoreInstancesFromString(char *inputstring, int maxposition = -1);

// -----------------------------------------------------------------------------
// Defmessage-handler functions
// -----------------------------------------------------------------------------

%section "Defmessage-handler functions"

unsigned FindDefmessageHandler(void *defclass,char *handlername,char *handlertype);
void     GetDefmessageHandlerList(void *defclass, DATA_OBJECT *result, int includeinherited);
char    *GetDefmessageHandlerName(void *defclass,unsigned index);
char    *GetDefmessageHandlerPPForm(void *defclass,unsigned index);
char    *GetDefmessageHandlerType(void *defclass,unsigned index);
int      GetDefmessageHandlerWatch(void *defclass,unsigned index);
unsigned GetNextDefmessageHandler(void *defclass,unsigned index);
int      IsDefmessageHandlerDeletable(void *defclass,unsigned index);
void     ListDefmessageHandlers(void *logname, void *defclass, int includeinherit);
int      UndefmessageHandler(void *defclass,unsigned index);
void     PreviewSend(char *logname, void *defclassPtr, char *messageName);
void     SetDefmessageHandlerWatch(int newstate, void *defclass, unsigned index);

// -----------------------------------------------------------------------------
// Definstances functions
// -----------------------------------------------------------------------------

%section "Definstances functions"

char    *DefinstancesModule(void *definstances);
void    *FindDefinstances(char *name);
void     GetDefinstancesList(DATA_OBJECT *result, void *theModule);
char    *GetDefinstancesName(void *definstances);
char    *GetDefinstancesPPForm(void *definstances);
void    *GetNextDefinstances(void *definstances);
int      IsDefinstancesDeletable(void *definstances);
void     ListDefinstances(char *logname, void *theModule);
BOOLEAN  Undefinstances(void *definstances);

// -----------------------------------------------------------------------------
// Defmodule functions
// -----------------------------------------------------------------------------

%section "Defmodule functions"

void    *FindDefmodule(char *defmodulename);
void    *GetCurrentModule();
void     GetDefmoduleList(DATA_OBJECT *result);
char    *GetDefmoduleName(void *defmodule);
char    *GetDefmodulePPForm(void *defmodule);
void    *GetNextDefmodule(void *defmodule);
void     ListDefmodules(char *logname);
void    *SetCurrentModule(void *defmodule);

// -----------------------------------------------------------------------------
// I/O Router functions
// -----------------------------------------------------------------------------

%section "I/O Router functions"

void ExitRouter(int exitcode);

int  GetcRouter(char *logname);
int  PrintRouter(char *logname, char *str);
int  UngetcRouter(int ch, char *logname);
int  ActivateRouter(char *routername);
int  DeactivateRouter(char *routername);
int  DeleteRouter(char *routername);

 
// AddRouter support. Allows a router to be implemented in Tcl

#ifdef SWIGTCL
%{
   typedef struct swig_router {
     char       *routerName;
     int         priority;
     Tcl_Interp *interp;
     char       *query;
     char       *print;
     char       *getc;
     char       *ungetc;
     char       *exit;
     struct     swig_router *next;
   } swig_router;

   static swig_router *swig_routers = 0;
   static swig_router *swig_find_router(char *name) {
     swig_router *r = swig_routers;
     while (r) {
       if (strcmp(r->routerName,name) == 0) return r;
       r = r->next;
     }
     return r;
   }
   static int tcl_router_query(char *logname) {
     swig_router *r;
     char temp[256];
     r = swig_find_router(logname);
     if (!r) return 0;
     if (r->query) {
       sprintf(temp,"%s %s", r->query, logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return 0;
   }
   static int tcl_router_print(char *logname, char *str) {
     swig_router *r;
     char *temp;
     r = swig_find_router(logname);
     if (r->print) {
       int len = strlen(r->print) + strlen(logname) + strlen(str) + 16;
       temp = (char *) malloc(len);
       sprintf(temp,"%s %s {%s}",r->print,logname,str);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 free(temp);
	 return 1;
       }
       free(temp);
     }
     return 0;
   }
   static int tcl_router_getc(char *logname) {
     char temp[256];
     swig_router *r;
     r = swig_find_router(logname);
     if (r->getc) {
       sprintf(temp,"%s %s",r->getc,logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return EOF;
   }
   static int tcl_router_ungetc(int ch, char *logname) {
     swig_router *r;
     char temp[256];
     r = swig_find_router(logname);
     if (r->ungetc){
       sprintf(temp,"%s %c %s",r->ungetc, ch, logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return EOF;
   }
   static int tcl_router_exit(int exitcode) {
     char temp[256];
     swig_router *r = swig_routers;
     while (r) {
       if (r->exit) {
	 sprintf(temp,"%s %d",r->exit,exitcode);
	 if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) return 1;
       }
       r = r->next;
     }
     return 0;
   }

   int MyAddRouter(Tcl_Interp *interp, char *routername, int priority,char *queryfunc, char *printfunc,
		   char *getcfunc, char *ungetcfunc, char *exitfunc) {

     
     int (*qfunc)(char *) = tcl_router_query;
     int (*pfunc)(char *, char *) = tcl_router_print;
     int (*cfunc)(char *) = tcl_router_getc;
     int (*ufunc)(int, char *)= tcl_router_ungetc;
     int (*efunc)(int) = tcl_router_exit;
     swig_router *r;

     if (!queryfunc) queryfunc = "";
     if (!printfunc) printfunc = "";
     if (!getcfunc) getcfunc = "";
     if (!ungetcfunc) ungetcfunc = "";
     if (!exitfunc) exitfunc = "";

     r = swig_find_router(routername);
     if (!r) {
       r = (swig_router *) malloc(sizeof(swig_router));
       r->routerName = 0;
       r->query = 0;
       r->print = 0;
       r->getc = 0;
       r->ungetc = 0;
       r->exit = 0;
       r->next = swig_routers;
       swig_routers = r;
     }
     if (r->routerName) free(r->routerName);
     if (r->query) free(r->query);
     if (r->print) free(r->print);
     if (r->getc) free(r->getc);
     if (r->ungetc) free(r->ungetc);
     if (r->exit) free(r->exit);
	
     if (strlen(queryfunc)) {
       r->query = (char *) malloc(strlen(queryfunc)+1);
       strcpy(r->query,queryfunc);
     } else {
       qfunc = 0;
       r->query = 0;
     }
     if (strlen(printfunc)) {
       r->print = (char *) malloc(strlen(printfunc)+1);
       strcpy(r->print,printfunc);
     } else {
       pfunc = 0;
       r->print = 0;
     }
     
     if (strlen(getcfunc)) {
       r->getc = (char *) malloc(strlen(getcfunc)+1);
       strcpy(r->getc,getcfunc);
     } else {
       cfunc = 0;
       r->getc = 0;
     }
    
     if (strlen(ungetcfunc)) {
       r->ungetc = (char *) malloc(strlen(ungetcfunc)+1);
       strcpy(r->ungetc,ungetcfunc);
     } else {
       ufunc = 0;
       r->ungetc = 0;
     }
     
     if (strlen(exitfunc)) {
       r->exit = (char *) malloc(strlen(exitfunc)+1);
       strcpy(r->exit,exitfunc);
     } else {
       efunc = 0;
       r->exit = 0;
     }
     r->routerName = (char *) malloc(strlen(routername)+1);
     strcpy(r->routerName,routername);
     r->priority = priority;
     r->interp = interp;
     return AddRouter(routername,priority,qfunc,pfunc,cfunc,ufunc,efunc);
   }
   %}

%name(AddRouter) int MyAddRouter(Tcl_Interp *interp, char *logname, int priority,char *query, char *print, char *getc, char *ungetc, char *exitfunc);

#endif




// -----------------------------------------------------------------------------
// Memory functions
// -----------------------------------------------------------------------------

%section "Memory Management"

int GetConserveMemory();
long MemRequests();
long MemUsed();
long ReleaseMem(int howmuch, int printmessage);
int  SetConserveMemory(int value);

%typedef int (*OutOfMemoryFunction)();

#ifdef SWIGTCL
%{
static  char *memoryproc = 0;
static  Tcl_Interp *memory_interp = 0;
int tcl_memoryproc(unsigned long size) {
  char temp[256];
  int res = 1;
  if (memoryproc && memory_interp) {
    sprintf(temp,"%s %d", memoryproc, size);
    Tcl_GlobalEval(memory_interp,temp);
    res = atoi(memory_interp->result);
  }
  return res;
}

tclint MySetOutOfMemoryFunction(Tcl_Interp *interp, char *memoryfunc) {
  char *oldmemoryfunc = memoryproc;
  memoryproc = (char *) malloc(strlen(memoryfunc)+8);
  strcpy(memoryproc, memoryfunc);
  memory_interp = interp;
  SetOutOfMemoryFunction(tcl_memoryproc);
  if (oldmemoryfunc) {
    Tcl_SetResult(interp,oldmemoryfunc,TCL_VOLATILE);
    free(oldmemoryfunc);
  } else {
    Tcl_SetResult(interp,"",TCL_STATIC);
  }
  return TCL_OK;
}
%}

%name(SetOutOfMemoryFunction) tclint MySetOutOfMemoryFunction(Tcl_Interp *interp, char *memoryfunc);
#endif


%section "User Defined Functions"

%{

  typedef struct tcl_clips_func {
      Tcl_Interp *interp;
      char *name;
      char *tclname;
      struct tcl_clips_func *next;
  } tcl_clips_func;
  
  static tcl_clips_func  *funcs = 0;
  
  static void clipstclfunc(DATA_OBJECT *returnvalue) {
    char temp1[64],temp2[512];
    char *invokedname;
    tcl_clips_func *f;

    /* This is a *MASSIVE* hack. Not documented. Just remember to duck and cover. */
    invokedname = ExpressionFunctionCallName(CurrentExpression)->contents;
    f = funcs;
    while (f) {
      if (strcmp(f->name,invokedname) == 0) {
	SWIG_MakePtr(temp1,returnvalue,"_DATA_OBJECT_p");
	sprintf(temp2,"%s %s",f->tclname,temp1);
	if (Tcl_GlobalEval(f->interp,temp2) != TCL_OK) {
	  SetpType(returnvalue,STRING);
	  SetpValue(returnvalue,AddSymbol(f->interp->result));
	}
	return;
      }
      f = f->next;
    }
    return;
  }

  int tcl_DefineFunction(Tcl_Interp *interp, char *name, char *tclname) {
    tcl_clips_func *f;
    f = (tcl_clips_func *) malloc(sizeof(tcl_clips_func));
    f->interp = interp;
    f->name = (char *) malloc(strlen(name)+1);
    strcpy(f->name,name);
    f->tclname = (char *) malloc(strlen(tclname)+1);
    strcpy(f->tclname,tclname);
    f->next = funcs;
    funcs = f;
    return DefineFunction(name,'u',PTIF clipstclfunc, tclname);
  }
  int tcl_DefineFunction2(Tcl_Interp *interp, char *name, char *tclname, char *restrictions) {
    tcl_clips_func *f;
    f = (tcl_clips_func *) malloc(sizeof(tcl_clips_func));
    f->interp = interp;
    f->name = (char *) malloc(strlen(name)+1);
    strcpy(f->name,name);
    f->tclname = (char *) malloc(strlen(tclname)+1);
    strcpy(f->tclname,tclname);
    f->next = funcs;
    funcs = f;
    return DefineFunction2(name,'u',PTIF clipstclfunc, tclname, restrictions);
  }


%}

%name(DefineFunction) int tcl_DefineFunction(Tcl_Interp *interp, char *name, char *tclname);
%name(DefineFunction2) int tcl_DefineFunction2(Tcl_Interp *interp, char *name, char *tclname,
char *restrictions);

int RtnArgCount();
int ArgCountCheck(char *name, int restriction, int count);
int ArgRangeCheck(char *name, int min, int max);

char  *RtnLexeme(int argpos);
double RtnDouble(int argpos);
long   RtnLong(int argpos);
DATA_OBJECT *RtnUnknown(int argpos, DATA_OBJECT *d);
int    ArgTypeCheck(char *name, int argpos, int typeint, DATA_OBJECT *result);

// -----------------------------------------------------------------------------
// tcleval
// -----------------------------------------------------------------------------
%{
  static Tcl_Interp *tcleval_interp = 0;
  static void *clipstcleval() {
    DATA_OBJECT d;
    char *temp;
    void *ret;

    if (ArgCountCheck("tcleval",EXACTLY,1) == -1) {
      return AddSymbol("");
    }
    if (ArgTypeCheck("tcleval",1,STRING,&d) == 0) {
      return AddSymbol("");
    }
    temp = (char *) malloc(strlen(DOToString(d))+1);
    strcpy(temp,DOToString(d));
    Tcl_GlobalEval(tcleval_interp,temp);
    ret = AddSymbol(tcleval_interp->result);
    free(temp);
    return ret;
  }

/* User defined functions */

void UserFunctions() {
  DefineFunction2("tcleval",'s', PTIF clipstcleval, "clipstcleval", "11s");
}
 %}

%init %{
  tcleval_interp = interp;   
%}


// ---------------------------------------------------------------------------
// Export all symbols in the clips namespace and declare a package
// ---------------------------------------------------------------------------

%init %{
#if (TCL_MAJOR_VERSION >= 8)
	 Tcl_Eval(interp,"namespace eval clips { namespace export * }");
#endif
    Tcl_PkgProvide(interp,"clips","1.0");
%}


