 
// AddRouter support. Allows a router to be implemented in Tcl

#ifdef SWIGTCL
%{
   typedef struct swig_router {
     char       *routerName;
     int         priority;
     Tcl_Interp *interp;
     char       *query;
     char       *print;
     char       *getc;
     char       *ungetc;
     char       *exit;
     struct     swig_router *next;
   } swig_router;

   static swig_router *swig_routers = 0;
   static swig_router *swig_find_router(char *name) {
     swig_router *r = swig_routers;
     while (r) {
       if (strcmp(r->routerName,name) == 0) return r;
       r = r->next;
     }
     return r;
   }
   static int tcl_router_query(char *logname) {
     swig_router *r;
     char temp[256];
     r = swig_find_router(logname);
     if (!r) return 0;
     if (r->query) {
       sprintf(temp,"%s %s", r->query, logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return 0;
   }
   static int tcl_router_print(char *logname, char *str) {
     swig_router *r;
     char *temp;
     r = swig_find_router(logname);
     if (r->print) {
       int len = strlen(r->print) + strlen(logname) + strlen(str) + 16;
       temp = (char *) malloc(len);
       sprintf(temp,"%s %s {%s}",r->print,logname,str);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 free(temp);
	 return 1;
       }
       free(temp);
     }
     return 0;
   }
   static int tcl_router_getc(char *logname) {
     char temp[256];
     swig_router *r;
     r = swig_find_router(logname);
     if (r->getc) {
       sprintf(temp,"%s %s",r->getc,logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return EOF;
   }
   static int tcl_router_ungetc(int ch, char *logname) {
     swig_router *r;
     char temp[256];
     r = swig_find_router(logname);
     if (r->ungetc){
       sprintf(temp,"%s %c %s",r->ungetc, ch, logname);
       if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) {
	 return atoi(r->interp->result);
       }
     }
     return EOF;
   }
   static int tcl_router_exit(int exitcode) {
     char temp[256];
     swig_router *r = swig_routers;
     while (r) {
       if (r->exit) {
	 sprintf(temp,"%s %d",r->exit,exitcode);
	 if (Tcl_GlobalEval(r->interp,temp) == TCL_OK) return 1;
       }
       r = r->next;
     }
     return 0;
   }

   int MyAddRouter(Tcl_Interp *interp, char *routername, int priority,char *queryfunc, char *printfunc,
		   char *getcfunc, char *ungetcfunc, char *exitfunc) {

     
     int (*qfunc)(char *) = tcl_router_query;
     int (*pfunc)(char *, char *) = tcl_router_print;
     int (*cfunc)(char *) = tcl_router_getc;
     int (*ufunc)(int, char *)= tcl_router_ungetc;
     int (*efunc)(int) = tcl_router_exit;
     swig_router *r;
     r = swig_find_router(routername);
     if (!r) {
       r = (swig_router *) malloc(sizeof(swig_router));
       r->next = swig_routers;
       swig_routers = r;
     }
     if (r->routerName) free(r->routerName);
     if (r->query) free(r->query);
     if (r->print) free(r->query);
     if (r->getc) free(r->getc);
     if (r->ungetc) free(r->ungetc);
     if (r->exit) free(r->exit);
     
     if (strlen(queryfunc)) {
       r->query = (char *) malloc(strlen(queryfunc)+1);
       strcpy(r->query,queryfunc);
     } else {
       qfunc = 0;
       r->query = 0;
     }
     if (strlen(printfunc)) {
       r->print = (char *) malloc(strlen(printfunc)+1);
       strcpy(r->print,printfunc);
     } else {
       pfunc = 0;
       r->print = 0;
     }
     
     if (strlen(getcfunc)) {
       r->getc = (char *) malloc(strlen(getcfunc)+1);
       strcpy(r->getc,getcfunc);
     } else {
       cfunc = 0;
       r->getc = 0;
     }
    
     if (strlen(ungetcfunc)) {
       r->ungetc = (char *) malloc(strlen(ungetcfunc)+1);
       strcpy(r->ungetc,ungetcfunc);
     } else {
       ufunc = 0;
       r->ungetc = 0;
     }
     
     if (strlen(exitfunc)) {
       r->exit = (char *) malloc(strlen(exitfunc)+1);
       strcpy(r->exit,exitfunc);
     } else {
       efunc = 0;
       r->exit = 0;
     }
     r->routerName = (char *) malloc(strlen(routername)+1);
     strcpy(r->routerName,routername);
     r->priority = priority;
     r->interp = interp;
     return AddRouter(routername,priority,qfunc,pfunc,cfunc,ufunc,efunc);
   }
   %}

%name(AddRouter) int MyAddRouter(Tcl_Interp *interp, char *logname, int priority,char *query, char *print, char *getc, char *ungetc, char *exitfunc);

#endif
